package se.jernlas.swing.gem;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

/**
 * Adapter class, adapting a Document (the model for most text based swing
 * components) to appear as an AnalyzableText.
 * 
 * @author jernlas
 * @see javax.swing.text.Document
 */
public class AnalyzableDocumentAdapter implements AnalyzableText {

   /**
    * The backing Document that is used as a source for change events and text
    * content. The document may not be changed once set by the constructor,
    * since this would most likely result in unexpected behaviour.
    */
   private final Document           doc;

   /**
    * A list of TextAnalyzers observing this AnalyzableText.
    */
   private final List<TextAnalyzer> analyzers = new ArrayList<TextAnalyzer>();

   /**
    * Creates a new AnalyzableDocument, wrapping around the supplied Document
    * object.
    * 
    * @param doc
    *           The document to use.
    */
   public AnalyzableDocumentAdapter(Document doc) {

      if (doc == null) {
         throw new IllegalArgumentException("The document may not be NULL.");
      }
      this.doc = doc;

      /* Use an anonymous subclass of DocumentListener to capture changes in the 
       * underlying Document and refire them as analyseText.
       * 
       */
      doc.addDocumentListener(new DocumentListener() {
         @Override
         public void changedUpdate(DocumentEvent e) {
            fireAnalyseTextEvent();
         }

         @Override
         public void insertUpdate(DocumentEvent e) {
            fireAnalyseTextEvent();
         }

         @Override
         public void removeUpdate(DocumentEvent e) {
            fireAnalyseTextEvent();
         }
      });
   }

   @Override
   public void addAnalyzer(TextAnalyzer a) {
      analyzers.add(a);
   }

   /**
    * Fires analyzeText for all TextAnalyzers observing this class. The list of
    * observers is defensively copied before calling so that no errors will
    * occur if the observer removes itself from the original list during 
    * execution of analyzeText().
    */
   private void fireAnalyseTextEvent() {
      List<TextAnalyzer> tmpList = Collections.unmodifiableList(analyzers);
      for (TextAnalyzer textAnalyzer : tmpList) {
         textAnalyzer.analyzeText(this);
      }

   }
   
   @Override
   public String getText() {
      try {
         return doc.getText(0, doc.getLength());
      } catch (BadLocationException e) {
         e.printStackTrace();
         return "";
      }
   }

   @Override
   public void removeAnalyzer(TextAnalyzer a) {
      analyzers.remove(a);
   }

}
