package se.jernlas.swing.gem;
import java.awt.Color;
import java.awt.Component;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JWindow;
import javax.swing.border.BevelBorder;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;
import javax.swing.text.Document;

import se.jernlas.swing.gem.types.TextType;

/**
 * A swing-component that helps the user to create text-documents 
 * by analyzing what the user writes and presenting helpful suggestions.
 * 
 * @author jernlas
 *
 */
public class Gem extends JComponent implements TextAnalyzer, AncestorListener {


   private static final long    serialVersionUID = -1449696234370220313L;
   
   /**
    * Last shown suggestions.
    */
   private final Set<TextType> lastShownSuggestions = new HashSet<TextType>();
   
   /**
    * The types of document to try and identify.
    */
   private final List<TextType> types            = new ArrayList<TextType>();
   
   /**
    * The window containing the list of suggestions
    */
   private JWindow              tipWindow                = null;
   
   /**
    * Indicates whether there are tips to display 
    */
   private volatile boolean     hasTips          = false;
   
   /**
    * The color of the gem when there are no tips available
    */
   private Color                inactiveColor    = Color.YELLOW;
   
   /**
    * The color of the gem when there are tips available
    */
   private Color                activeColor      = Color.GREEN;

   /**
    * Creates a default gem without associating it with a text.
    * This can be done later by calling addAnalyzer() on the text.
    */
   public Gem() {
      addAncestorListener(this);
      setOpaque(false);
   }

   /**
    * Creates a default gem and registers it as an analyzer for the supplied text.
    * @param text The text to register as an analyzer for.
    */
   public Gem(AnalyzableText text) {
      this();
      text.addAnalyzer(this);
   }

   /**
    * Convenience constructor. 
    * Same as the one accepting a AnalyzableText parameter, 
    * only it accepts a Document, which is wrapped in an AnalyzableDocumentAdapter
    * to make it analyzable.
    * @param document The document to analyze.
    */
   public Gem(Document document) {
      this(new AnalyzableDocumentAdapter(document));
   }

   @Override
   public void addTextType(TextType t) {
      types.add(t);
   }

   @Override
   public void analyzeText(AnalyzableText text) {
      final String currentText = text.getText();
      List<TextType> matchingTypes = new LinkedList<TextType>();
      for (TextType t : types) {
         if (t.isThisType(currentText)) {
            matchingTypes.add(t);
         }
      }
      hasTips = (matchingTypes.size() > 0);
      if (hasTips) {
         displaySuggestions(matchingTypes);
      } else {
         hideSuggestions();
      }

      repaint();
   }

   @Override
   public void ancestorAdded(AncestorEvent event) {
      hideSuggestions();
   }

   @Override
   public void ancestorMoved(AncestorEvent event) {
      if (event.getSource() != tipWindow) {
         placePopup();
      }
   }

   @Override
   public void ancestorRemoved(AncestorEvent event) {
      hideSuggestions();
   }

   /**
    * Displays a list of Suggestions.
    * @param matchingTypes The TextTypes that matches the text.
    */
   private void displaySuggestions(Collection<TextType> matchingTypes) {
      

      boolean hasAll = lastShownSuggestions.containsAll(matchingTypes);
      boolean noExtra = lastShownSuggestions.size() == matchingTypes.size();
      
      // If displaying the same suggestions, don't redraw.
      if(hasAll && noExtra){
         return;
      } else {
         lastShownSuggestions.clear();
         lastShownSuggestions.addAll(matchingTypes);
      }
      
      // If there is already a window, dispose of it.
      if (tipWindow != null) {
         tipWindow.dispose();
      }

      // Create a new window and populate it with components generated 
      // from the selected TextTypes.
      tipWindow = new JWindow(getFrame(null));
      JPanel p = new JPanel();
      p.setLayout(new BoxLayout(p, BoxLayout.PAGE_AXIS));
      tipWindow.setContentPane(p);
      p.setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));
      for (TextType textType : matchingTypes) {
         p.add(textType.getTip());
      }
      final JButton b = new JButton("No thanks...");
      b.addActionListener(new ActionListener(){

         @Override
         public void actionPerformed(ActionEvent e) {
            hideSuggestions();
         }
         
      });
      p.add(b);
     
      // Make the pop-up window behave as part of the component.
      getFrame(null).addWindowFocusListener(new WindowAdapter() {
         @Override
         public void windowGainedFocus(WindowEvent evt) {
            if (tipWindow != null && !lastShownSuggestions.isEmpty()) {
               tipWindow.setVisible(true);
            }
         }

         @Override
         public void windowLostFocus(WindowEvent evt) {
            if (tipWindow != null && lastShownSuggestions.isEmpty()) {
               tipWindow.setVisible(false);
            }
         }

      });
      
      // Pack, show and position the pop-up window
      tipWindow.pack();
      tipWindow.toFront();
      tipWindow.setVisible(true);
      placePopup();
      tipWindow.requestFocusInWindow();
   }

   /**
    * Recursively gets the root Frame.
    * @param c The component to start from, if null this is used.
    * @return The root Frame.
    */
   private Frame getFrame(Component c) {
      if (c == null) {
         c = this;
      }
      if (c.getParent() instanceof Frame) {
         return (Frame) c.getParent();
      }
      return getFrame(c.getParent());
   }

   /**
    * Disables the pop-up suggestions.
    */
   private void hideSuggestions() {
      if (tipWindow != null) {
         tipWindow.setVisible(false);
         lastShownSuggestions.clear();
      }

   }

   @Override
   public void paint(Graphics g) {
      if (isOpaque()) {
         g.setColor(getBackground());
         g.drawRect(0, 0, getWidth() - 1, getHeight() - 1);
      }
      
      if (hasTips) {
         g.setColor(activeColor);
      } else {
         g.setColor(inactiveColor);
      }
      g.fillOval(0, 0, getWidth() - 1, getHeight() - 1);
      
      g.setColor(getForeground());
      g.drawOval(0, 0, getWidth() - 1, getHeight() - 1);
      
      g.setColor(Color.WHITE);
      g.fillOval((int)(0.7*getWidth()), (int)(0.25*getHeight()), (int)(0.15*(getWidth()-1)), (int)(0.15*(getHeight()-1)));
      
   }

   /**
    * Correctly position the pop-up, depending on the position of the main component.
    */
   private void placePopup() {
      if (tipWindow != null) {
         Point pt = this.getLocationOnScreen();
         pt.translate(0, this.getHeight());
         tipWindow.setLocation(pt);
      }
   }

   @Override
   public void removeTextType(TextType t) {
      types.remove(t);
   }

}
